# arepytools/io/parsing/pyxb_metadata.py
# -*- coding: utf-8 -*-
# PyXB bindings for NM:e92452c8d3e28a9e27abfc9994d2007779e7f4c9
# Generated 2021-02-17 11:26:26.605622 by PyXB version 1.2.6 using Python 3.8.5.final.0
# Namespace AbsentNamespace0

from __future__ import unicode_literals
import pyxb
import pyxb.binding
import pyxb.binding.saxer
import io
import pyxb.utils.utility
import pyxb.utils.domutils
import sys
import pyxb.utils.six as _six
# Unique identifier for bindings created at the same time
_GenerationUID = pyxb.utils.utility.UniqueIdentifier('urn:uuid:972c4fe8-710a-11eb-bb6b-7085c2d17c63')

# Version of PyXB used to generate the bindings
_PyXBVersion = '1.2.6'
# Generated bindings are not compatible across PyXB versions
if pyxb.__version__ != _PyXBVersion:
    raise pyxb.PyXBVersionError(_PyXBVersion)

# A holder for module-level binding classes so we can access them from
# inside class definitions where property names may conflict.
_module_typeBindings = pyxb.utils.utility.Object()

# Import bindings for namespaces imported into schema
import pyxb.binding.datatypes
from . import pyxb_metadata_types as _ImportedBinding_pyxb_metadata_types

# NOTE: All namespace declarations are reserved within the binding
Namespace = pyxb.namespace.CreateAbsentNamespace()
Namespace.configureCategories(['typeBinding', 'elementBinding'])

def CreateFromDocument (xml_text, default_namespace=None, location_base=None):
    """Parse the given XML and use the document element to create a
    Python instance.

    @param xml_text An XML document.  This should be data (Python 2
    str or Python 3 bytes), or a text (Python 2 unicode or Python 3
    str) in the L{pyxb._InputEncoding} encoding.

    @keyword default_namespace The L{pyxb.Namespace} instance to use as the
    default namespace where there is no default namespace in scope.
    If unspecified or C{None}, the namespace of the module containing
    this function will be used.

    @keyword location_base: An object to be recorded as the base of all
    L{pyxb.utils.utility.Location} instances associated with events and
    objects handled by the parser.  You might pass the URI from which
    the document was obtained.
    """

    if pyxb.XMLStyle_saxer != pyxb._XMLStyle:
        dom = pyxb.utils.domutils.StringToDOM(xml_text)
        return CreateFromDOM(dom.documentElement, default_namespace=default_namespace)
    if default_namespace is None:
        default_namespace = Namespace.fallbackNamespace()
    saxer = pyxb.binding.saxer.make_parser(fallback_namespace=default_namespace, location_base=location_base)
    handler = saxer.getContentHandler()
    xmld = xml_text
    if isinstance(xmld, _six.text_type):
        xmld = xmld.encode(pyxb._InputEncoding)
    saxer.parse(io.BytesIO(xmld))
    instance = handler.rootObject()
    return instance

def CreateFromDOM (node, default_namespace=None):
    """Create a Python instance from the given DOM node.
    The node tag must correspond to an element declaration in this module.

    @deprecated: Forcing use of DOM interface is unnecessary; use L{CreateFromDocument}."""
    if default_namespace is None:
        default_namespace = Namespace.fallbackNamespace()
    return pyxb.binding.basis.element.AnyCreateFromDOM(node, default_namespace)


# Complex type AresysXmlDocType with content type ELEMENT_ONLY
class AresysXmlDocType (pyxb.binding.basis.complexTypeDefinition):
    """Complex type AresysXmlDocType with content type ELEMENT_ONLY"""
    _TypeDefinition = None
    _ContentTypeTag = pyxb.binding.basis.complexTypeDefinition._CT_ELEMENT_ONLY
    _Abstract = False
    _ExpandedName = pyxb.namespace.ExpandedName(Namespace, 'AresysXmlDocType')
    _XSDLocation = pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 5, 2)
    _ElementMap = {}
    _AttributeMap = {}
    # Base type is pyxb.binding.datatypes.anyType
    
    # Element NumberOfChannels uses Python identifier NumberOfChannels
    __NumberOfChannels = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'NumberOfChannels'), 'NumberOfChannels', '__AbsentNamespace0_AresysXmlDocType_NumberOfChannels', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 7, 6), )

    
    NumberOfChannels = property(__NumberOfChannels.value, __NumberOfChannels.set, None, None)

    
    # Element VersionNumber uses Python identifier VersionNumber
    __VersionNumber = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'VersionNumber'), 'VersionNumber', '__AbsentNamespace0_AresysXmlDocType_VersionNumber', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 8, 6), )

    
    VersionNumber = property(__VersionNumber.value, __VersionNumber.set, None, None)

    
    # Element Description uses Python identifier Description
    __Description = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'Description'), 'Description', '__AbsentNamespace0_AresysXmlDocType_Description', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 9, 6), )

    
    Description = property(__Description.value, __Description.set, None, None)

    
    # Element Channel uses Python identifier Channel
    __Channel = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'Channel'), 'Channel', '__AbsentNamespace0_AresysXmlDocType_Channel', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 11, 8), )

    
    Channel = property(__Channel.value, __Channel.set, None, None)

    _ElementMap.update({
        __NumberOfChannels.name() : __NumberOfChannels,
        __VersionNumber.name() : __VersionNumber,
        __Description.name() : __Description,
        __Channel.name() : __Channel
    })
    _AttributeMap.update({
        
    })
_module_typeBindings.AresysXmlDocType = AresysXmlDocType
Namespace.addCategoryObject('typeBinding', 'AresysXmlDocType', AresysXmlDocType)


# Complex type ChannelType with content type EMPTY
class ChannelType (_ImportedBinding_pyxb_metadata_types.TreeElementBaseType):
    """Complex type ChannelType with content type EMPTY"""
    _TypeDefinition = None
    _ContentTypeTag = pyxb.binding.basis.complexTypeDefinition._CT_EMPTY
    _Abstract = False
    _ExpandedName = pyxb.namespace.ExpandedName(Namespace, 'ChannelType')
    _XSDLocation = pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 45, 2)
    _ElementMap = _ImportedBinding_pyxb_metadata_types.TreeElementBaseType._ElementMap.copy()
    _AttributeMap = _ImportedBinding_pyxb_metadata_types.TreeElementBaseType._AttributeMap.copy()
    # Base type is _ImportedBinding_pyxb_metadata_types.TreeElementBaseType
    
    # Attribute ContentID uses Python identifier ContentID
    __ContentID = pyxb.binding.content.AttributeUse(pyxb.namespace.ExpandedName(None, 'ContentID'), 'ContentID', '__AbsentNamespace0_ChannelType_ContentID', pyxb.binding.datatypes.string)
    __ContentID._DeclarationLocation = pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 48, 8)
    __ContentID._UseLocation = pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 48, 8)
    
    ContentID = property(__ContentID.value, __ContentID.set, None, None)

    
    # Attribute Number inherited from {aresysTypes}TreeElementBaseType
    
    # Attribute Total inherited from {aresysTypes}TreeElementBaseType
    _ElementMap.update({
        
    })
    _AttributeMap.update({
        __ContentID.name() : __ContentID
    })
_module_typeBindings.ChannelType = ChannelType
Namespace.addCategoryObject('typeBinding', 'ChannelType', ChannelType)


# Complex type [anonymous] with content type ELEMENT_ONLY
class CTD_ANON (ChannelType):
    """Complex type [anonymous] with content type ELEMENT_ONLY"""
    _TypeDefinition = None
    _ContentTypeTag = pyxb.binding.basis.complexTypeDefinition._CT_ELEMENT_ONLY
    _Abstract = False
    _ExpandedName = None
    _XSDLocation = pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 12, 10)
    _ElementMap = ChannelType._ElementMap.copy()
    _AttributeMap = ChannelType._AttributeMap.copy()
    # Base type is ChannelType
    
    # Element RasterInfo uses Python identifier RasterInfo
    __RasterInfo = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'RasterInfo'), 'RasterInfo', '__AbsentNamespace0_CTD_ANON_RasterInfo', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 16, 18), )

    
    RasterInfo = property(__RasterInfo.value, __RasterInfo.set, None, None)

    
    # Element ROI uses Python identifier ROI
    __ROI = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'ROI'), 'ROI', '__AbsentNamespace0_CTD_ANON_ROI', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 17, 18), )

    
    ROI = property(__ROI.value, __ROI.set, None, None)

    
    # Element DataSetInfo uses Python identifier DataSetInfo
    __DataSetInfo = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'DataSetInfo'), 'DataSetInfo', '__AbsentNamespace0_CTD_ANON_DataSetInfo', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 18, 18), )

    
    DataSetInfo = property(__DataSetInfo.value, __DataSetInfo.set, None, None)

    
    # Element SwathInfo uses Python identifier SwathInfo
    __SwathInfo = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'SwathInfo'), 'SwathInfo', '__AbsentNamespace0_CTD_ANON_SwathInfo', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 19, 18), )

    
    SwathInfo = property(__SwathInfo.value, __SwathInfo.set, None, None)

    
    # Element SamplingConstants uses Python identifier SamplingConstants
    __SamplingConstants = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'SamplingConstants'), 'SamplingConstants', '__AbsentNamespace0_CTD_ANON_SamplingConstants', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 20, 18), )

    
    SamplingConstants = property(__SamplingConstants.value, __SamplingConstants.set, None, None)

    
    # Element AcquisitionTimeLine uses Python identifier AcquisitionTimeLine
    __AcquisitionTimeLine = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'AcquisitionTimeLine'), 'AcquisitionTimeLine', '__AbsentNamespace0_CTD_ANON_AcquisitionTimeLine', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 21, 18), )

    
    AcquisitionTimeLine = property(__AcquisitionTimeLine.value, __AcquisitionTimeLine.set, None, None)

    
    # Element DataStatistics uses Python identifier DataStatistics
    __DataStatistics = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'DataStatistics'), 'DataStatistics', '__AbsentNamespace0_CTD_ANON_DataStatistics', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 22, 18), )

    
    DataStatistics = property(__DataStatistics.value, __DataStatistics.set, None, None)

    
    # Element BurstInfo uses Python identifier BurstInfo
    __BurstInfo = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'BurstInfo'), 'BurstInfo', '__AbsentNamespace0_CTD_ANON_BurstInfo', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 23, 18), )

    
    BurstInfo = property(__BurstInfo.value, __BurstInfo.set, None, None)

    
    # Element StateVectorData uses Python identifier StateVectorData
    __StateVectorData = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'StateVectorData'), 'StateVectorData', '__AbsentNamespace0_CTD_ANON_StateVectorData', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 24, 18), )

    
    StateVectorData = property(__StateVectorData.value, __StateVectorData.set, None, None)

    
    # Element DopplerCentroid uses Python identifier DopplerCentroid
    __DopplerCentroid = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'DopplerCentroid'), 'DopplerCentroid', '__AbsentNamespace0_CTD_ANON_DopplerCentroid', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 25, 18), )

    
    DopplerCentroid = property(__DopplerCentroid.value, __DopplerCentroid.set, None, None)

    
    # Element DopplerRate uses Python identifier DopplerRate
    __DopplerRate = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'DopplerRate'), 'DopplerRate', '__AbsentNamespace0_CTD_ANON_DopplerRate', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 26, 18), )

    
    DopplerRate = property(__DopplerRate.value, __DopplerRate.set, None, None)

    
    # Element TopsAzimuthModulationRate uses Python identifier TopsAzimuthModulationRate
    __TopsAzimuthModulationRate = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'TopsAzimuthModulationRate'), 'TopsAzimuthModulationRate', '__AbsentNamespace0_CTD_ANON_TopsAzimuthModulationRate', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 27, 18), )

    
    TopsAzimuthModulationRate = property(__TopsAzimuthModulationRate.value, __TopsAzimuthModulationRate.set, None, None)

    
    # Element SlantToGround uses Python identifier SlantToGround
    __SlantToGround = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'SlantToGround'), 'SlantToGround', '__AbsentNamespace0_CTD_ANON_SlantToGround', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 28, 18), )

    
    SlantToGround = property(__SlantToGround.value, __SlantToGround.set, None, None)

    
    # Element GroundToSlant uses Python identifier GroundToSlant
    __GroundToSlant = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'GroundToSlant'), 'GroundToSlant', '__AbsentNamespace0_CTD_ANON_GroundToSlant', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 29, 18), )

    
    GroundToSlant = property(__GroundToSlant.value, __GroundToSlant.set, None, None)

    
    # Element SlantToIncidence uses Python identifier SlantToIncidence
    __SlantToIncidence = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'SlantToIncidence'), 'SlantToIncidence', '__AbsentNamespace0_CTD_ANON_SlantToIncidence', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 30, 18), )

    
    SlantToIncidence = property(__SlantToIncidence.value, __SlantToIncidence.set, None, None)

    
    # Element SlantToElevation uses Python identifier SlantToElevation
    __SlantToElevation = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'SlantToElevation'), 'SlantToElevation', '__AbsentNamespace0_CTD_ANON_SlantToElevation', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 31, 18), )

    
    SlantToElevation = property(__SlantToElevation.value, __SlantToElevation.set, None, None)

    
    # Element AttitudeInfo uses Python identifier AttitudeInfo
    __AttitudeInfo = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'AttitudeInfo'), 'AttitudeInfo', '__AbsentNamespace0_CTD_ANON_AttitudeInfo', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 32, 18), )

    
    AttitudeInfo = property(__AttitudeInfo.value, __AttitudeInfo.set, None, None)

    
    # Element GroundCornerPoints uses Python identifier GroundCornerPoints
    __GroundCornerPoints = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'GroundCornerPoints'), 'GroundCornerPoints', '__AbsentNamespace0_CTD_ANON_GroundCornerPoints', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 33, 18), )

    
    GroundCornerPoints = property(__GroundCornerPoints.value, __GroundCornerPoints.set, None, None)

    
    # Element Pulse uses Python identifier Pulse
    __Pulse = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'Pulse'), 'Pulse', '__AbsentNamespace0_CTD_ANON_Pulse', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 34, 18), )

    
    Pulse = property(__Pulse.value, __Pulse.set, None, None)

    
    # Element CoregPoly uses Python identifier CoregPoly
    __CoregPoly = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'CoregPoly'), 'CoregPoly', '__AbsentNamespace0_CTD_ANON_CoregPoly', True, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 35, 18), )

    
    CoregPoly = property(__CoregPoly.value, __CoregPoly.set, None, None)

    
    # Element AntennaInfo uses Python identifier AntennaInfo
    __AntennaInfo = pyxb.binding.content.ElementDeclaration(pyxb.namespace.ExpandedName(None, 'AntennaInfo'), 'AntennaInfo', '__AbsentNamespace0_CTD_ANON_AntennaInfo', False, pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 36, 18), )

    
    AntennaInfo = property(__AntennaInfo.value, __AntennaInfo.set, None, None)

    
    # Attribute ContentID inherited from ChannelType
    
    # Attribute Number inherited from {aresysTypes}TreeElementBaseType
    
    # Attribute Total inherited from {aresysTypes}TreeElementBaseType
    _ElementMap.update({
        __RasterInfo.name() : __RasterInfo,
        __ROI.name() : __ROI,
        __DataSetInfo.name() : __DataSetInfo,
        __SwathInfo.name() : __SwathInfo,
        __SamplingConstants.name() : __SamplingConstants,
        __AcquisitionTimeLine.name() : __AcquisitionTimeLine,
        __DataStatistics.name() : __DataStatistics,
        __BurstInfo.name() : __BurstInfo,
        __StateVectorData.name() : __StateVectorData,
        __DopplerCentroid.name() : __DopplerCentroid,
        __DopplerRate.name() : __DopplerRate,
        __TopsAzimuthModulationRate.name() : __TopsAzimuthModulationRate,
        __SlantToGround.name() : __SlantToGround,
        __GroundToSlant.name() : __GroundToSlant,
        __SlantToIncidence.name() : __SlantToIncidence,
        __SlantToElevation.name() : __SlantToElevation,
        __AttitudeInfo.name() : __AttitudeInfo,
        __GroundCornerPoints.name() : __GroundCornerPoints,
        __Pulse.name() : __Pulse,
        __CoregPoly.name() : __CoregPoly,
        __AntennaInfo.name() : __AntennaInfo
    })
    _AttributeMap.update({
        
    })
_module_typeBindings.CTD_ANON = CTD_ANON


AresysXmlDoc = pyxb.binding.basis.element(pyxb.namespace.ExpandedName(Namespace, 'AresysXmlDoc'), AresysXmlDocType, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 4, 2))
Namespace.addCategoryObject('elementBinding', AresysXmlDoc.name().localName(), AresysXmlDoc)



AresysXmlDocType._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'NumberOfChannels'), pyxb.binding.datatypes.unsignedInt, scope=AresysXmlDocType, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 7, 6)))

AresysXmlDocType._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'VersionNumber'), pyxb.binding.datatypes.double, scope=AresysXmlDocType, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 8, 6)))

AresysXmlDocType._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'Description'), pyxb.binding.datatypes.string, scope=AresysXmlDocType, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 9, 6)))

AresysXmlDocType._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'Channel'), CTD_ANON, scope=AresysXmlDocType, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 11, 8)))

def _BuildAutomaton ():
    # Remove this helper function from the namespace after it is invoked
    global _BuildAutomaton
    del _BuildAutomaton
    import pyxb.utils.fac as fac

    counters = set()
    cc_0 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 10, 6))
    counters.add(cc_0)
    states = []
    final_update = None
    symbol = pyxb.binding.content.ElementUse(AresysXmlDocType._UseForTag(pyxb.namespace.ExpandedName(None, 'NumberOfChannels')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 7, 6))
    st_0 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_0)
    final_update = None
    symbol = pyxb.binding.content.ElementUse(AresysXmlDocType._UseForTag(pyxb.namespace.ExpandedName(None, 'VersionNumber')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 8, 6))
    st_1 = fac.State(symbol, is_initial=False, final_update=final_update, is_unordered_catenation=False)
    states.append(st_1)
    final_update = set()
    symbol = pyxb.binding.content.ElementUse(AresysXmlDocType._UseForTag(pyxb.namespace.ExpandedName(None, 'Description')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 9, 6))
    st_2 = fac.State(symbol, is_initial=False, final_update=final_update, is_unordered_catenation=False)
    states.append(st_2)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_0, False))
    symbol = pyxb.binding.content.ElementUse(AresysXmlDocType._UseForTag(pyxb.namespace.ExpandedName(None, 'Channel')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 11, 8))
    st_3 = fac.State(symbol, is_initial=False, final_update=final_update, is_unordered_catenation=False)
    states.append(st_3)
    transitions = []
    transitions.append(fac.Transition(st_1, [
         ]))
    st_0._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_2, [
         ]))
    st_1._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_3, [
         ]))
    st_2._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_3, [
        fac.UpdateInstruction(cc_0, True) ]))
    st_3._set_transitionSet(transitions)
    return fac.Automaton(states, counters, False, containing_state=None)
AresysXmlDocType._Automaton = _BuildAutomaton()




CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'RasterInfo'), _ImportedBinding_pyxb_metadata_types.RasterInfoType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 16, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'ROI'), _ImportedBinding_pyxb_metadata_types.ROIType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 17, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'DataSetInfo'), _ImportedBinding_pyxb_metadata_types.DataSetInfoType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 18, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'SwathInfo'), _ImportedBinding_pyxb_metadata_types.SwathInfoType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 19, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'SamplingConstants'), _ImportedBinding_pyxb_metadata_types.SamplingConstantsType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 20, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'AcquisitionTimeLine'), _ImportedBinding_pyxb_metadata_types.AcquisitionTimelineType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 21, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'DataStatistics'), _ImportedBinding_pyxb_metadata_types.DataStatisticsType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 22, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'BurstInfo'), _ImportedBinding_pyxb_metadata_types.BurstInfoType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 23, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'StateVectorData'), _ImportedBinding_pyxb_metadata_types.StateVectorDataType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 24, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'DopplerCentroid'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 25, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'DopplerRate'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 26, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'TopsAzimuthModulationRate'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 27, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'SlantToGround'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 28, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'GroundToSlant'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 29, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'SlantToIncidence'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 30, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'SlantToElevation'), _ImportedBinding_pyxb_metadata_types.polyType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 31, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'AttitudeInfo'), _ImportedBinding_pyxb_metadata_types.AttitudeInfoType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 32, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'GroundCornerPoints'), _ImportedBinding_pyxb_metadata_types.GroundCornersPointsType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 33, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'Pulse'), _ImportedBinding_pyxb_metadata_types.PulseType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 34, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'CoregPoly'), _ImportedBinding_pyxb_metadata_types.polyCoregType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 35, 18)))

CTD_ANON._AddElement(pyxb.binding.basis.element(pyxb.namespace.ExpandedName(None, 'AntennaInfo'), _ImportedBinding_pyxb_metadata_types.AntennaInfoType, scope=CTD_ANON, location=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 36, 18)))

def _BuildAutomaton_ ():
    # Remove this helper function from the namespace after it is invoked
    global _BuildAutomaton_
    del _BuildAutomaton_
    import pyxb.utils.fac as fac

    counters = set()
    cc_0 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 16, 18))
    counters.add(cc_0)
    cc_1 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 17, 18))
    counters.add(cc_1)
    cc_2 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 18, 18))
    counters.add(cc_2)
    cc_3 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 19, 18))
    counters.add(cc_3)
    cc_4 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 20, 18))
    counters.add(cc_4)
    cc_5 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 21, 18))
    counters.add(cc_5)
    cc_6 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 22, 18))
    counters.add(cc_6)
    cc_7 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 23, 18))
    counters.add(cc_7)
    cc_8 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 24, 18))
    counters.add(cc_8)
    cc_9 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 25, 18))
    counters.add(cc_9)
    cc_10 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 26, 18))
    counters.add(cc_10)
    cc_11 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 27, 18))
    counters.add(cc_11)
    cc_12 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 28, 18))
    counters.add(cc_12)
    cc_13 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 29, 18))
    counters.add(cc_13)
    cc_14 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 30, 18))
    counters.add(cc_14)
    cc_15 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 31, 18))
    counters.add(cc_15)
    cc_16 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 32, 18))
    counters.add(cc_16)
    cc_17 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 33, 18))
    counters.add(cc_17)
    cc_18 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 34, 18))
    counters.add(cc_18)
    cc_19 = fac.CounterCondition(min=0, max=None, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 35, 18))
    counters.add(cc_19)
    cc_20 = fac.CounterCondition(min=0, max=1, metadata=pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 36, 18))
    counters.add(cc_20)
    states = []
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_0, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'RasterInfo')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 16, 18))
    st_0 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_0)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_1, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'ROI')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 17, 18))
    st_1 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_1)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_2, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'DataSetInfo')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 18, 18))
    st_2 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_2)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_3, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'SwathInfo')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 19, 18))
    st_3 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_3)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_4, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'SamplingConstants')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 20, 18))
    st_4 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_4)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_5, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'AcquisitionTimeLine')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 21, 18))
    st_5 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_5)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_6, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'DataStatistics')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 22, 18))
    st_6 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_6)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_7, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'BurstInfo')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 23, 18))
    st_7 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_7)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_8, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'StateVectorData')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 24, 18))
    st_8 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_8)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_9, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'DopplerCentroid')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 25, 18))
    st_9 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_9)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_10, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'DopplerRate')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 26, 18))
    st_10 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_10)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_11, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'TopsAzimuthModulationRate')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 27, 18))
    st_11 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_11)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_12, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'SlantToGround')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 28, 18))
    st_12 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_12)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_13, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'GroundToSlant')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 29, 18))
    st_13 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_13)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_14, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'SlantToIncidence')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 30, 18))
    st_14 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_14)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_15, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'SlantToElevation')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 31, 18))
    st_15 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_15)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_16, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'AttitudeInfo')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 32, 18))
    st_16 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_16)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_17, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'GroundCornerPoints')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 33, 18))
    st_17 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_17)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_18, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'Pulse')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 34, 18))
    st_18 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_18)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_19, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'CoregPoly')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 35, 18))
    st_19 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_19)
    final_update = set()
    final_update.add(fac.UpdateInstruction(cc_20, False))
    symbol = pyxb.binding.content.ElementUse(CTD_ANON._UseForTag(pyxb.namespace.ExpandedName(None, 'AntennaInfo')), pyxb.utils.utility.Location('aresys_generic_metadata.xsd', 36, 18))
    st_20 = fac.State(symbol, is_initial=True, final_update=final_update, is_unordered_catenation=False)
    states.append(st_20)
    transitions = []
    transitions.append(fac.Transition(st_0, [
        fac.UpdateInstruction(cc_0, True) ]))
    transitions.append(fac.Transition(st_1, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_2, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_3, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_4, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_5, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_0, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_0, False) ]))
    st_0._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_1, [
        fac.UpdateInstruction(cc_1, True) ]))
    transitions.append(fac.Transition(st_2, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_3, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_4, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_5, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_1, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_1, False) ]))
    st_1._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_2, [
        fac.UpdateInstruction(cc_2, True) ]))
    transitions.append(fac.Transition(st_3, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_4, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_5, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_2, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_2, False) ]))
    st_2._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_3, [
        fac.UpdateInstruction(cc_3, True) ]))
    transitions.append(fac.Transition(st_4, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_5, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_3, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_3, False) ]))
    st_3._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_4, [
        fac.UpdateInstruction(cc_4, True) ]))
    transitions.append(fac.Transition(st_5, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_4, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_4, False) ]))
    st_4._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_5, [
        fac.UpdateInstruction(cc_5, True) ]))
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_5, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_5, False) ]))
    st_5._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_6, [
        fac.UpdateInstruction(cc_6, True) ]))
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_6, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_6, False) ]))
    st_6._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_7, [
        fac.UpdateInstruction(cc_7, True) ]))
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_7, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_7, False) ]))
    st_7._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_8, [
        fac.UpdateInstruction(cc_8, True) ]))
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_8, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_8, False) ]))
    st_8._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_9, [
        fac.UpdateInstruction(cc_9, True) ]))
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_9, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_9, False) ]))
    st_9._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_10, [
        fac.UpdateInstruction(cc_10, True) ]))
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_10, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_10, False) ]))
    st_10._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_11, [
        fac.UpdateInstruction(cc_11, True) ]))
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_11, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_11, False) ]))
    st_11._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_12, [
        fac.UpdateInstruction(cc_12, True) ]))
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_12, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_12, False) ]))
    st_12._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_13, [
        fac.UpdateInstruction(cc_13, True) ]))
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_13, False) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_13, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_13, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_13, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_13, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_13, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_13, False) ]))
    st_13._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_14, [
        fac.UpdateInstruction(cc_14, True) ]))
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_14, False) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_14, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_14, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_14, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_14, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_14, False) ]))
    st_14._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_15, [
        fac.UpdateInstruction(cc_15, True) ]))
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_15, False) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_15, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_15, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_15, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_15, False) ]))
    st_15._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_16, [
        fac.UpdateInstruction(cc_16, True) ]))
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_16, False) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_16, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_16, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_16, False) ]))
    st_16._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_17, [
        fac.UpdateInstruction(cc_17, True) ]))
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_17, False) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_17, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_17, False) ]))
    st_17._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_18, [
        fac.UpdateInstruction(cc_18, True) ]))
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_18, False) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_18, False) ]))
    st_18._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_19, [
        fac.UpdateInstruction(cc_19, True) ]))
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_19, False) ]))
    st_19._set_transitionSet(transitions)
    transitions = []
    transitions.append(fac.Transition(st_20, [
        fac.UpdateInstruction(cc_20, True) ]))
    st_20._set_transitionSet(transitions)
    return fac.Automaton(states, counters, True, containing_state=None)
CTD_ANON._Automaton = _BuildAutomaton_()




# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import os
import logging
import collections
from shapely.geometry import MultiPoint
from biopal.utility.utility_functions import getBinaryNameFromChannelIDX
from biopal.io.xml_io import (
    parse_input_file,
    geographic_boundaries_latlon,
    write_input_file,
    stack_based_processing,
)
from biopal.utility.utility_functions import Task, start_logging
from biopal.io.data_io import readBiomassHeader_core
from arepytools.timing.precisedatetime import PreciseDateTime


class InvalidInputError(ValueError):
    pass


class dataset_query(Task):
    """Query over input dataSet
    
    Performs a query on the BioPAL input dataSet: the query rule is specified 
    into the input file "dataset_query" section to get stacks from a 
    specified geographical and temporal region.
    Writes in output an updated input file, filled with new section 
    "stack_based_processing"

    Methods
    -------
    run( input_file_path )
        run the dataSet query
    name : str
        name of the APP

    Examples
    --------
    To ececute it, use its run method after initialization:

    >>> from biopal.dataset_query.dataset_query import dataset_query
    >>> dq_obj = dataset_query()
    >>> input_file_up = dq_obj.run( input_file )

    - input_file: path of the input file containing "dataset_query" section
    - input_file_up: path of the input file, updated with "stack_based_processing" section
    """

    def __init__(self):
        super().__init__()

    def _run(self, input_file):

        if isinstance(input_file, str):
            input_params_obj = parse_input_file(input_file)
        else:
            input_params_obj = input_file
        if (
            input_params_obj.L2_product is None
            or input_params_obj.output_specification is None
            or input_params_obj.dataset_query is None
        ):

            error_message = [
                "BioPAL APP input file should contain at least"
                ' "L2_product", "output_specification" and "dataset_query" sections'
            ]
            raise InvalidInputError(error_message)

        if not hasattr(logging.getLoggerClass().root.handlers[0], "baseFilename"):
            start_logging(
                input_params_obj.output_specification.output_folder,
                input_params_obj.L2_product,
                "DEBUG",
                app_name="dataset_query",
            )
        (
            stack_composition,
            geographic_boundaries,
            geographic_boundaries_per_stack,
        ) = data_select_by_date_and_boundaries(input_params_obj.dataset_query)
        if not len(stack_composition):
            logging.error(
                "Cannot find any data which meets the conditions to run the processing: see the log warnings. \n"
            )
            logging.info("BIOMASS L2 Processor ended without any processing performed.")
            return 0

        logging.info("Geographic Boundaries extracted [deg]:")
        logging.info("Latitude  min: " + str(geographic_boundaries.lat_min))
        logging.info("Latitude  max: " + str(geographic_boundaries.lat_max))
        logging.info("Longitude min: " + str(geographic_boundaries.lon_min))
        logging.info("Longitude max: " + str(geographic_boundaries.lon_max))

        # write the input files for the activated chains:
        stack_based_proc_obj = fill_stack_based_processing_obj(
            input_params_obj, stack_composition, geographic_boundaries, geographic_boundaries_per_stack,
        )

        input_file_up_xml = os.path.join(
            input_params_obj.output_specification.output_folder,
            "InputFile_StackBasedProcessing{}.xml".format(input_params_obj.L2_product),
        )
        input_params_obj.output_specification.output_folder = os.path.join(
            input_params_obj.output_specification.output_folder, input_params_obj.L2_product,
        )
        input_params_obj.stack_based_processing = stack_based_proc_obj
        write_input_file(input_params_obj, input_file_up_xml)

        end_message = "Query completed."
        logging.info(end_message)
        print(end_message)

        return input_file_up_xml


def data_select_by_date_and_boundaries(dataset_query_obj):

    geographic_boundaries_curr = geographic_boundaries_latlon(None, None, None, None)

    # from all the database of data, selectonly the ones which matches the
    # input parameters and return a dictionary with the stacks to be used

    start_date = dataset_query_obj.L1C_date.start
    stop_date = dataset_query_obj.L1C_date.stop

    # the stack_composition will contain Stack_IDs and Stack Compositions for the available dataset at desired dates and boundaries
    stacks_composition = {}
    # stack_composition_dates_match and stack_composition_boundaries_match are used only for warning display
    # in case no data is found in the desired dates+boudaries, but just for one of the two
    stacks_composition_only_dates_match = {}
    stacks_composition_only_boundaries_match = {}
    stacks_composition_not_enough_acq = {}
    stacks_geographic_boundaries = {}

    # initialize the list which will contain all the header_biomass extracted realizations

    # cycle over all the data set and read the binary header: we read just first channel, no need for all of them
    channel_idx = 1
    pf_names_list = os.listdir(dataset_query_obj.L1C_repository)

    for pf_idx, pf_name in enumerate(pf_names_list):
        date_is_matching = False
        boundaries_are_matching = False
        stack_is_matching = False
        num_acq_are_enough = False

        if len(pf_name) != 47:
            error_msg = 'Unrecognized uniqie acquisition ID (it will be skipped): "{}" from repository {}; the correct format is "GC_XX_H_XXX.XX_RGSW_XX_RGSBSW_X_AZSW_XX_BSL_XX" where each "X" is an integer'.format(
                pf_name, dataset_query_obj.L1C_repository
            )
            logging.error(error_msg)
            continue

        output_pf_full_name = os.path.join(dataset_query_obj.L1C_repository, pf_name)

        bin_file_name = getBinaryNameFromChannelIDX(pf_name, channel_idx)
        raster_file = os.path.join(output_pf_full_name, bin_file_name)
        if not os.path.exists(raster_file):
            logging.warning(
                'The folder "'
                + bin_file_name
                + '" inside L1C repository is NOT a product folder and it will be skipped \n'
            )
            continue

        # read the current header
        (
            datestr,
            lon_min,
            lon_max,
            lat_min,
            lat_max,
            unique_acq_id,
            resolution_m_slant_rg,
            resolution_m_az,
            sensor_velocity,
        ) = readBiomassHeader_core(raster_file)

        uniqie_stack_id = unique_acq_id[0:40]

        # first test:check the date:
        date = PreciseDateTime().set_from_utc_string(datestr)

        if pf_idx == 0:
            edge_values = {
                "date_min": date,
                "date_max": date,
                "lon_min": lon_min,
                "lon_max": lon_max,
                "lat_min": lat_min,
                "lat_max": lat_max,
            }

        else:
            edge_values["date_min"] = min(edge_values["date_min"], date)
            edge_values["date_max"] = max(edge_values["date_max"], date)
            edge_values["lon_min"] = min(edge_values["lon_min"], lon_min)
            edge_values["lon_max"] = max(edge_values["lon_max"], lon_max)
            edge_values["lat_min"] = min(edge_values["lat_min"], lat_min)
            edge_values["lat_max"] = max(edge_values["lat_max"], lat_max)

        if uniqie_stack_id in stacks_geographic_boundaries.keys():

            stacks_geographic_boundaries[uniqie_stack_id].lon_min = min(
                stacks_geographic_boundaries[uniqie_stack_id].lon_min, lon_min
            )
            stacks_geographic_boundaries[uniqie_stack_id].lon_max = max(
                stacks_geographic_boundaries[uniqie_stack_id].lon_max, lon_max
            )
            stacks_geographic_boundaries[uniqie_stack_id].lat_min = min(
                stacks_geographic_boundaries[uniqie_stack_id].lat_min, lat_min
            )
            stacks_geographic_boundaries[uniqie_stack_id].lat_max = max(
                stacks_geographic_boundaries[uniqie_stack_id].lat_max, lat_max
            )
        else:
            geographic_boundaries_per_stack = geographic_boundaries_latlon(None, None, None, None)
            geographic_boundaries_per_stack.lon_min = lon_min
            geographic_boundaries_per_stack.lon_max = lon_max
            geographic_boundaries_per_stack.lat_min = lat_min
            geographic_boundaries_per_stack.lat_max = lat_max
            stacks_geographic_boundaries[uniqie_stack_id] = geographic_boundaries_per_stack

        # check 2/4 search for data inside the user dates boundaries:
        date_is_matching = check_on_dates(start_date, stop_date, date)

        # check 3/4 search for data inside the user ground coordinates boundaries:
        boundaries_are_matching = check_on_boundaries(
            lon_min, lon_max, lat_min, lat_max, dataset_query_obj.boundaries.point
        )

        # check 4/4 (optional) ckeck if data matches one of the user specified stack:
        stack_is_matching = check_on_stacks_to_find(uniqie_stack_id, dataset_query_obj.stacks_to_find)

        if date_is_matching and boundaries_are_matching and stack_is_matching:  # and num_acq_are_enough:
            # store information for data that satisfy all previous checks:
            if not uniqie_stack_id in stacks_composition.keys():
                stacks_composition[uniqie_stack_id] = [unique_acq_id]
            else:
                stacks_composition[uniqie_stack_id].append(unique_acq_id)

            try:
                geographic_boundaries_curr.lon_min = min(geographic_boundaries_curr.lon_min, lon_min)
                geographic_boundaries_curr.lon_max = max(geographic_boundaries_curr.lon_max, lon_max)
                geographic_boundaries_curr.lat_min = min(geographic_boundaries_curr.lat_min, lat_min)
                geographic_boundaries_curr.lat_max = max(geographic_boundaries_curr.lat_max, lat_max)
            except:
                geographic_boundaries_curr.lon_min = lon_min
                geographic_boundaries_curr.lon_max = lon_max
                geographic_boundaries_curr.lat_min = lat_min
                geographic_boundaries_curr.lat_max = lat_max

        elif date_is_matching and stack_is_matching and num_acq_are_enough:
            # store information for warning display
            if not uniqie_stack_id in stacks_composition_only_dates_match.keys():
                stacks_composition_only_dates_match[uniqie_stack_id] = [unique_acq_id]
            else:
                stacks_composition_only_dates_match[uniqie_stack_id].append(unique_acq_id)

        elif boundaries_are_matching and stack_is_matching and num_acq_are_enough:
            # store information for warning display
            if not uniqie_stack_id in stacks_composition_only_boundaries_match.keys():
                stacks_composition_only_boundaries_match[uniqie_stack_id] = [unique_acq_id]
            else:
                stacks_composition_only_boundaries_match[uniqie_stack_id].append(unique_acq_id)

    # return the headers_list and some useful messages to the user
    num_stacks = len(stacks_composition)
    if num_stacks == 0:
        logging.warning("\n")
        logging.warning("Cannot find any valid data:")
        if len(stacks_composition_only_dates_match) or len(stacks_composition_only_boundaries_match):

            logging.warning("    be aware that data In L1cRepository are confined inside following boundaries:")
            logging.warning("    minimun date is       {}".format(edge_values["date_min"]))
            logging.warning("    maximun date is       {}".format(edge_values["date_max"]))
            logging.warning("    minimun Longitude is  {}".format(edge_values["lon_min"]))
            logging.warning("    maximum Longitude is  {}".format(edge_values["lon_max"]))
            logging.warning("    minimun Latitude  is  {}".format(edge_values["lat_min"]))
            logging.warning("    maximum Latitude  is  {} \n".format(edge_values["lat_max"]))
            logging.warning("    Instead the user asked for following boundaries:")
            logging.warning("    user min date         {}".format(start_date))
            logging.warning("    user max date         {}".format(stop_date))

            point_dict = dataset_query_obj.boundaries.point

            logging.warning("    user min Longitude    {}".format(min([x["Longitude"] for x in point_dict])))
            logging.warning("    user max Longitude    {}".format(max([x["Longitude"] for x in point_dict])))
            logging.warning("    user min Latitude     {}".format(min([x["Latitude"] for x in point_dict])))
            logging.warning("    user max Latitude     {} \n".format(max([x["Latitude"] for x in point_dict])))

            if len(stacks_composition_only_dates_match):
                logging.warning(
                    "    Following stacks are confined in the desired L1cDates but not in the desired geographic boundaries"
                )
                logging.warning("    {} \n".format(list(stacks_composition_only_dates_match.keys())))

            if len(stacks_composition_only_boundaries_match):
                logging.warning(
                    "    Following stacks are confined in the desired geographic boundaries but not in the desired L1cDates"
                )
                logging.warning("    {} \n".format(list(stacks_composition_only_boundaries_match.keys())))

        if len(stacks_composition_not_enough_acq):
            logging.warning(
                "Following stacks contain just ONE acquisition (a minimum of TWO are needed to have ONE baseline)"
            )
            logging.warning("{} \n".format(list(stacks_composition_not_enough_acq.keys())))

        if (
            not len(stacks_composition_only_dates_match)
            and not len(stacks_composition_only_boundaries_match)
            and not len(stacks_composition_not_enough_acq)
        ):
            logging.warning("Cannot find any data, the L1cRepository is empty")
            pf_idx = 0

    else:
        logging.info("Found #{} stacks: {}".format(num_stacks, list(stacks_composition.keys())))
        logging.info("Stacks composition: \n")
        for stack_curr, pf_list in stacks_composition.items():
            pf_list.sort()
            logging.info("Stack " + stack_curr + " composition ( unique acquisitions id ):")
            for pf_name in pf_list:
                logging.info("    " + pf_name)

    stacks_composition = collections.OrderedDict(sorted(stacks_composition.items()))

    return stacks_composition, geographic_boundaries_curr, stacks_geographic_boundaries


def fill_stack_based_processing_obj(
    input_params_obj, stack_composition, geographic_boundaries, geographic_boundaries_per_stack,
):

    """
    Internal function called by the dataset_query APP:
        
        after the query over the dataSet, the datset_query APP fills the structure
        to be written into the xml input file for the next APP, which is the 
        stack based processing.
        The output structure contains all the queried stacks paths and names from dataSet ( 
        and auxiliary paths), other than boundaries.
        The returned object "stack_based_processing_obj" structure will have the specific 
        ields needed by the activated chain 
        (AGB, FD, FH or TOMO_FH), other than all the common fields to each chain.

        Usage of the returned object:
            The returned object can be added to the input_params_obj and it 
            can be written to disk if needed, as (stack_based_processing_obj is 
                                                  overwritten by this command, if already present):
                - input_params_obj.stack_based_processing_obj = stack_based_processing_obj
                - write_input_file(input_params_obj, input_file_xml)
            
    """

    (
        ECEF_grid_file_names,
        kz_file_names,
        off_nadir_angle_file_names,
        reference_height_file_names,
        slant_range_distances_file_names,
        slope_file_names,
        forest_mask_catalogue_folder,
        dem_folder,
        reference_agb_folder,
        average_covariance_folder,
        system_decorrelation_fun_folder,
        calibration_screens_folder,
        forest_height_folder,
        calibration_screens_file_names,
    ) = get_auxiliary_paths(stack_composition, input_params_obj.dataset_query.L1C_aux_repository)

    if input_params_obj.L2_product == "AGB":
        system_decorrelation_fun_folder = None
        average_covariance_folder = None

    if input_params_obj.L2_product == "FD":
        reference_agb_folder = None
        system_decorrelation_fun_folder = None
        forest_height_folder = None

    if input_params_obj.L2_product == "FH":
        reference_agb_folder = None
        average_covariance_folder = None
        forest_height_folder = None

    if input_params_obj.L2_product == "TOMO_FH":

        stack_composition_TOMO = {}
        for stack_id, acquisitions_list in stack_composition.items():
            if len(acquisitions_list) >= 3:
                stack_composition_TOMO[stack_id] = stack_composition[stack_id]

        len_stacks_all = stack_composition.keys()
        len_stacks_tomo = stack_composition_TOMO.keys()
        if len_stacks_tomo:
            if len_stacks_tomo < len_stacks_all:
                logging.info(
                    "The stacks with less than #3 acquisitions cannot be used to perform TOMO FH estimation \
                     and they will be removed: #{} stacks among #{} totaL, have been removed".format(
                        len_stacks_tomo, len_stacks_all
                    )
                )
        else:
            error_message = (
                "cannot find any stack with more than #2 acquisitions: TOMO FH estimation cannot be executed"
            )
            logging.error(error_message)
            raise InvalidInputError(error_message)

        stack_composition = stack_composition_TOMO
        reference_agb_folder = None
        forest_mask_catalogue_folder = None
        system_decorrelation_fun_folder = None
        average_covariance_folder = None
        forest_height_folder = None

    stack_based_processing_obj = stack_based_processing(
        stack_composition,
        ECEF_grid_file_names,
        kz_file_names,
        slant_range_distances_file_names,
        off_nadir_angle_file_names,
        slope_file_names,
        reference_height_file_names,
        calibration_screens_file_names,
        dem_folder,
        reference_agb_folder,
        forest_mask_catalogue_folder,
        system_decorrelation_fun_folder,
        average_covariance_folder,
        forest_height_folder,
        geographic_boundaries,
        geographic_boundaries_per_stack,
    )

    return stack_based_processing_obj


def get_auxiliary_paths(stack_composition, L1c_aux_data_repository):
    # get the names of the folders containing of the auxiliary data: the names
    # shouls match the default ones

    ECEFGRID_folder = os.path.join(L1c_aux_data_repository, "Geometry", "ECEFGRID")
    KZ_folder = os.path.join(L1c_aux_data_repository, "Geometry", "KZ")
    off_nadir_angle_folder = os.path.join(L1c_aux_data_repository, "Geometry", "OffNadirAngles")
    reference_height_folder = os.path.join(L1c_aux_data_repository, "Geometry", "ReferenceHeight")
    slant_range_distances = os.path.join(L1c_aux_data_repository, "Geometry", "SlantRangeDistances")
    slope_folder = os.path.join(L1c_aux_data_repository, "Geometry", "Slope")

    forest_mask_catalogue_folder = os.path.join(L1c_aux_data_repository, "ForestMask")
    dem_folder = os.path.join(L1c_aux_data_repository, "DEM")
    reference_agb_folder = os.path.join(L1c_aux_data_repository, "ReferenceAGB")
    average_covariance_folder = os.path.join(L1c_aux_data_repository, "AverageCovariance")
    system_decorrelation_fun_folder = os.path.join(L1c_aux_data_repository, "SystemDecorrelationFunction")
    calibration_screens_folder = os.path.join(L1c_aux_data_repository, "CalibrationScreens")
    forest_height_folder = os.path.join(L1c_aux_data_repository, "ForestHeight")

    # Those are stack based auxiliary:
    ECEF_grid_file_names = {}
    kz_file_names = {}
    off_nadir_angle_file_names = {}
    reference_height_file_names = {}
    slant_range_distances_file_names = {}
    slope_file_names = {}
    calibration_screens_file_names = {}

    for stack_id in stack_composition.keys():

        ECEF_grid_file_names[stack_id] = os.path.join(r"" + ECEFGRID_folder, stack_id)
        kz_file_names[stack_id] = os.path.join(r"" + KZ_folder, stack_id)
        off_nadir_angle_file_names[stack_id] = os.path.join(r"" + off_nadir_angle_folder, stack_id)
        reference_height_file_names[stack_id] = os.path.join(r"" + reference_height_folder, stack_id)
        slant_range_distances_file_names[stack_id] = os.path.join(r"" + slant_range_distances, stack_id)
        slope_file_names[stack_id] = os.path.join(r"" + slope_folder, stack_id)
        calibration_screens_file_names[stack_id] = os.path.join(r"" + calibration_screens_folder, stack_id)

    return (
        ECEF_grid_file_names,
        kz_file_names,
        off_nadir_angle_file_names,
        reference_height_file_names,
        slant_range_distances_file_names,
        slope_file_names,
        forest_mask_catalogue_folder,
        dem_folder,
        reference_agb_folder,
        average_covariance_folder,
        system_decorrelation_fun_folder,
        calibration_screens_folder,
        forest_height_folder,
        calibration_screens_file_names,
    )


def check_on_dates(start_date, stop_date, date):
    # core function to check the dates match

    if date >= start_date and date <= stop_date:
        date_is_matching = True
    else:
        date_is_matching = False

    return date_is_matching


def check_on_boundaries(lon_min, lon_max, lat_min, lat_max, point_dict):
    # core function to check if the data falls inside the user geographic boundaries

    # create the geographic polygon containing the user input searching area:
    data_longitudes = [x["Longitude"] for x in point_dict]
    data_latitudes = [x["Latitude"] for x in point_dict]
    data_lat_lon_coords = tuple(zip(data_latitudes, data_longitudes))
    data_polygon_obj = MultiPoint(data_lat_lon_coords).convex_hull

    # create the geographic polygon containing the current data
    user_lat_lon_coords = tuple(zip([lat_min, lat_min, lat_max, lat_max], [lon_min, lon_max, lon_min, lon_max]))
    user_polygon_obj = MultiPoint(user_lat_lon_coords).convex_hull

    # check if the data polygon is at least partially intersected by the user polygon
    boundaries_are_matching = user_polygon_obj.intersects(data_polygon_obj)

    return boundaries_are_matching


def check_on_stacks_to_find(uniqie_stack_id, stacks_to_find):
    # core function: this is optional, in the sense that stacks_to_find can be empty, in this case don't check on stacks (set it to True)
    if len(stacks_to_find):
        if uniqie_stack_id in stacks_to_find:
            stack_is_matching = True
        else:
            stack_is_matching = False
    else:
        stack_is_matching = True

    return stack_is_matching



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Ellipsoid module
----------------
"""

import numpy as np


class Ellipsoid:
    """
    Ellipsoid class
    """

    def __init__(self, first_semi_axis, second_semi_axis):
        """Initialize the object with the specified ellipsoid parameters

        :param first_semi_axis: length of the first semi-axis of the ellipsoid
        :param second_semi_axis: length of the second semi-axis of the ellipsoid
        """

        self._semi_major_axis = float(max(first_semi_axis, second_semi_axis))
        self._semi_minor_axis = float(min(first_semi_axis, second_semi_axis))

        self._semi_axes_ratio_min_max = self._semi_minor_axis / self._semi_major_axis
        self._eccentricity_square = 1 - self._semi_axes_ratio_min_max ** 2
        self._eccentricity = np.sqrt(self._eccentricity_square)
        self._ep2 = 1. / self._semi_axes_ratio_min_max ** 2 - 1

    @property
    def semi_major_axis(self):
        """Semi-major axis length"""
        return self._semi_major_axis

    @property
    def semi_minor_axis(self):
        """Semi-minor axis length"""
        return self._semi_minor_axis

    @property
    def semi_axes_ratio_min_max(self):
        """Min/Max semi-axis ratio"""
        return self._semi_axes_ratio_min_max

    @property
    def eccentricity_square(self):
        """Squared eccentricity"""
        return self._eccentricity_square

    @property
    def eccentricity(self):
        """Eccentricity"""
        return self._eccentricity

    @property
    def ep2(self):
        """Ep2"""
        return self._ep2



# SPDX-FileCopyrightText: BioPAL <biopal@esa.int>
# SPDX-License-Identifier: MIT

import logging
import numpy as np
from scipy.sparse import csr_matrix
from scipy.signal import firwin, oaconvolve
from biopal.utility.constants import LIGHTSPEED


def main_covariance_estimation_SR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_slant_rg,
    pixel_spacing_az,
    incidence_angle_rad,
    carrier_frequency_hz,
    range_bandwidth_hz,
):
    """Covariance estimation:  inputs are in Slant Range radar coordinates
    see also main_covariance_estimation_GR"""

    # compute the pixels spacing in ground range, from slant range:
    pixel_spacing_grd_x = pixel_spacing_slant_rg / np.sin(incidence_angle_rad)
    pixel_spacing_grd_y = pixel_spacing_az

    slant_range_spacing_s = pixel_spacing_slant_rg / LIGHTSPEED

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_GR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_grd_x,
        pixel_spacing_grd_y,
        carrier_frequency_hz,
        range_bandwidth_hz,
        slant_range_spacing_s,
    )

    return MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def main_covariance_estimation_SSF_SR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_slant_rg,
    pixel_spacing_az,
    incidence_angle_rad,
    R,
    look_angles,
    ground_slope,
    carrier_frequency_hz,
    range_bandwidth_hz,
):
    """Covariance estimation:  inputs are in Slant Range radar coordinates
    see also main_covariance_estimation_GR"""

    # compute the pixels spacing in ground range, from slant range:
    pixel_spacing_grd_x = pixel_spacing_slant_rg / np.sin(incidence_angle_rad)
    pixel_spacing_grd_y = pixel_spacing_az

    slant_range_spacing_s = pixel_spacing_slant_rg / LIGHTSPEED

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_SSF_GR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_grd_x,
        pixel_spacing_grd_y,
        R,
        look_angles,
        ground_slope,
        carrier_frequency_hz,
        range_bandwidth_hz,
        slant_range_spacing_s,
    )

    return MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def main_covariance_estimation_GR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_grd_x,
    pixel_spacing_grd_y,
    carrier_frequency_hz,
    range_bandwidth_hz,
    slant_range_spacing_s=None,
):
    """Covariance estimation:  inputs are in Ground Range coordinates
    see also main_covariance_estimation_SR"""

    cov_est_opt_str = covariance_options_struct_filler(
        data_stack,
        cov_est_window_size,
        pixel_spacing_grd_x,
        pixel_spacing_grd_y,
        carrier_frequency_hz,
        range_bandwidth_hz,
        slant_range_spacing_s,
    )

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = MPMBCovarianceEstimation(
        data_stack, cov_est_opt_str
    )

    return MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def main_covariance_estimation_SSF_GR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_grd_x,
    pixel_spacing_grd_y,
    R,
    look_angles,
    ground_slope,
    carrier_frequency_hz,
    range_bandwidth_hz,
    slant_range_spacing_s=None,
):
    """Covariance estimation:  inputs are in Ground Range coordinates
    see also main_covariance_estimation_SR"""

    cov_est_opt_str = covariance_options_struct_filler(
        data_stack,
        cov_est_window_size,
        pixel_spacing_grd_x,
        pixel_spacing_grd_y,
        carrier_frequency_hz,
        range_bandwidth_hz,
        slant_range_spacing_s,
    )

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = MPMBCovarianceEstimationSSF(
        data_stack, cov_est_opt_str, R, look_angles, ground_slope
    )

    return MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def covariance_options_struct_filler(
    data_stack,
    cov_est_window_size,
    pixel_spacing_grd_x,
    pixel_spacing_grd_y,
    carrier_frequency_hz,
    range_bandwidth_hz,
    slant_range_spacing_s=None,
):

    first_level_key = next(iter(data_stack.keys()))

    if first_level_key == "hh" or first_level_key == "hv" or first_level_key == "vh" or first_level_key == "vv":
        # notched stack, has not acquisitions dict
        pol_names = list(data_stack.keys())
        Num_of_pols = len(pol_names)

    else:
        acq_names = list(data_stack.keys())
        first_acq_dict = data_stack[acq_names[0]]
        pol_names = list(first_acq_dict.keys())
        Num_of_pols = len(pol_names)

    class cov_est_opt_str:
        pass

    cov_est_opt_str.NW_r = np.int64(cov_est_window_size / pixel_spacing_grd_x / 2) * 2 + 1
    cov_est_opt_str.NW_a = np.int64(cov_est_window_size / pixel_spacing_grd_y / 2) * 2 + 1

    cov_est_opt_str.NWmono_r = np.floor((cov_est_opt_str.NW_r - 1) / 2).astype("int64")
    cov_est_opt_str.NWmono_a = np.floor((cov_est_opt_str.NW_a - 1) / 2).astype("int64")

    cov_est_opt_str.subs_F_r = cov_est_opt_str.NWmono_r
    cov_est_opt_str.subs_F_a = cov_est_opt_str.NWmono_a

    cov_est_opt_str.polarimetric_mask = np.ones((Num_of_pols, Num_of_pols))

    cov_est_opt_str.wavelenght = LIGHTSPEED / carrier_frequency_hz
    cov_est_opt_str.B = range_bandwidth_hz
    cov_est_opt_str.f0 = carrier_frequency_hz
    cov_est_opt_str.dt = slant_range_spacing_s

    return cov_est_opt_str


def main_correlation_estimation_SR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_slant_rg,
    pixel_spacing_az,
    incidence_angle_rad,
    carrier_frequency_hz,
    range_bandwidth_hz,
):

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_SR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_slant_rg,
        pixel_spacing_az,
        incidence_angle_rad,
        carrier_frequency_hz,
        range_bandwidth_hz,
    )

    # Normalizing covariance matrix
    MPMB_correlation, norm_diag = Covariance4D2Correlation4D(MPMB_covariance)

    return MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def main_correlation_estimation_SSF_SR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_slant_rg,
    pixel_spacing_az,
    incidence_angle_rad,
    R,
    look_angles,
    ground_slope,
    carrier_frequency_hz,
    range_bandwidth_hz,
):

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_SSF_SR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_slant_rg,
        pixel_spacing_az,
        incidence_angle_rad,
        R,
        look_angles,
        ground_slope,
        carrier_frequency_hz,
        range_bandwidth_hz,
    )

    # Normalizing covariance matrix
    MPMB_correlation, norm_diag = Covariance4D2Correlation4D(MPMB_covariance)

    return MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def main_correlation_estimation_GR(
    data_stack, cov_est_window_size, pixel_spacing_grd_x, pixel_spacing_grd_y, carrier_frequency_hz, range_bandwidth_hz,
):

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_GR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_grd_x,
        pixel_spacing_grd_y,
        carrier_frequency_hz,
        range_bandwidth_hz,
    )

    # Normalizing covariance matrix
    MPMB_correlation, norm_diag = Covariance4D2Correlation4D(MPMB_covariance)

    return MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def main_correlation_estimation_SSF_GR(
    data_stack,
    cov_est_window_size,
    pixel_spacing_grd_x,
    pixel_spacing_grd_y,
    R,
    look_angles,
    ground_slope,
    carrier_frequency_hz,
    range_bandwidth_hz,
):

    (MPMB_covariance, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a,) = main_covariance_estimation_SSF_GR(
        data_stack,
        cov_est_window_size,
        pixel_spacing_grd_x,
        pixel_spacing_grd_y,
        R,
        look_angles,
        ground_slope,
        carrier_frequency_hz,
        range_bandwidth_hz,
    )

    # Normalizing covariance matrix
    MPMB_correlation, norm_diag = Covariance4D2Correlation4D(MPMB_covariance)

    return MPMB_correlation, rg_vec_subs, az_vec_subs, subs_F_r, subs_F_a


def MPMBCovarianceEstimation(D_in, opt_str):
    """MPMBCovarianceEstimation

    INPUTS:
        D_in: can be a:
                  dictionary with acquisition keys (Num_Baselines) each containing a dictionary with hh hv vh and vv
                  polarization keys containing arrays with dimensions [Nrg x Naz]
            or (in case of ground notched input) :
                  dictionary with hh hv vh and vv
                  polarization keys containing arrays with dimensions [Nrg x Naz]
        opt_str:
            NWmono_r: One-sided range average window [pixel]
            subs_F_r: Range subsampling factor
            NWmono_a: One-sided azimuth average window [pixel]
            subs_F_a: Azimuth subsampling factor
            polarimetric_mask: (optional) [Npol x Npol] logical matrix
                                          stating the polarization combination to
                                          be computed

    OUTPUTS:
        Cov_MPMB: Multi-polarimetric multi-baseline covariance matrix
        rg_out: range subsampled axis
        az_out: azimuth subsampled axis
    """

    first_level_key = next(iter(D_in.keys()))

    if first_level_key == "hh" or first_level_key == "hv" or first_level_key == "vh" or first_level_key == "vv":
        # notched stack, has not acquisitions dict, so we need to encapsulate the input in an external dummy dictionary
        D = {"single_acq": D_in}
    else:
        D = D_in
    del D_in

    acq_names = list(D.keys())
    N_imm = len(acq_names)
    first_acq_dict = D[acq_names[0]]
    pol_names = list(first_acq_dict.keys())
    num_pols = len(pol_names)
    N_rg, N_az = first_acq_dict[pol_names[0]].shape
    cell_data = True
    double_data = first_acq_dict[pol_names[0]].dtype == "complex128"

    try:
        Mask = opt_str.polarimetric_mask
    except AttributeError:
        Mask = np.ones((num_pols, num_pols))

    # Range filter matrix
    class filtering_matrix_opt_str:
        pass

    filtering_matrix_opt_str.Nin = N_rg
    filtering_matrix_opt_str.NWmono = opt_str.NWmono_r
    filtering_matrix_opt_str.subs_F = opt_str.subs_F_r
    Fr, rg_out, Rnorm = build_filtering_matrix(filtering_matrix_opt_str)
    N_rg_out = rg_out.size
    Fr_normalized = Rnorm @ Fr

    # Azimuth filter matrix
    filtering_matrix_opt_str.Nin = N_az
    filtering_matrix_opt_str.NWmono = opt_str.NWmono_a
    filtering_matrix_opt_str.subs_F = opt_str.subs_F_a
    Fa, az_out, Anorm = build_filtering_matrix(filtering_matrix_opt_str)
    N_az_out = az_out.size
    Fa_normalized_transposed = (Anorm @ Fa).T

    # Init
    Cov_MPMB = np.zeros((num_pols * N_imm, num_pols * N_imm, N_rg_out, N_az_out), dtype=np.complex64)

    logging.info("    MPMB covariance estimation...")
    # nan_mask = np.zeros( D[0].shape , dtype=bool) N_rg, N_az
    nan_mask = np.zeros((N_rg, N_az, N_imm), dtype=bool)

    for pol_name_curr in pol_names:
        all_acq_data = np.zeros((N_rg, N_az, N_imm), dtype=complex)
        for acq_idx_curr, acq_curr in enumerate(acq_names):
            all_acq_data[:, :, acq_idx_curr] = D[acq_curr][pol_name_curr]
        nan_mask = nan_mask + np.isnan(all_acq_data)

    # this has been moved from the innner cycle (referring to PM9)
    for acq_idx_curr, acq_curr in enumerate(acq_names):
        for pol_name_curr in pol_names:
            D[acq_curr][pol_name_curr][nan_mask[:, :, acq_idx_curr]] = 0

    for ch_i, pol_id_i in enumerate(pol_names):

        ind_i = np.arange(N_imm) + ch_i * N_imm
        for ch_j in np.arange(ch_i, num_pols):
            pol_id_j = pol_names[ch_j]
            if Mask[ch_i, ch_j] == 1:
                print("    .")
                ind_j = np.arange(N_imm) + ch_j * N_imm
                I = np.zeros((N_imm, N_imm, N_rg_out, N_az_out), dtype=np.complex64)
                for n_idx, acq_id_n in enumerate(acq_names):
                    if ch_i == ch_j:
                        m_min = n_idx  # ?
                    else:
                        m_min = 0

                    for m_idx in np.arange(m_min, N_imm):
                        acq_id_m = acq_names[m_idx]
                        if cell_data:
                            if double_data:
                                temp = D[acq_id_n][pol_id_i] * np.conjugate(D[acq_id_m][pol_id_j])
                            else:
                                temp = D[acq_id_n][pol_id_i].astype("complex128") * np.conjugate(
                                    D[acq_id_m][pol_id_j]
                                ).astype("complex128")
                        else:
                            if double_data:
                                temp = D[acq_id_n][pol_id_i] * np.conjugate(D[acq_id_m][pol_id_j])
                            else:
                                temp = D[acq_id_n][pol_id_i].astype("complex128") * np.conjugate(
                                    D[acq_id_m][pol_id_j]
                                ).astype("complex128")
                        temp = Fr_normalized @ temp
                        temp = temp @ Fa_normalized_transposed
                        I[n_idx, m_idx, :, :] = temp.astype("complex64")
                Cov_MPMB[ind_i[:, np.newaxis], ind_j[np.newaxis, :], :, :] = I

    # Symmetric part generation
    diag_mask = np.tile(
        (np.eye(num_pols * N_imm).reshape(num_pols * N_imm, num_pols * N_imm, 1, 1)) > 0, [1, 1, N_rg_out, N_az_out],
    )
    Cov_MPMB = Cov_MPMB + np.conjugate(np.moveaxis(Cov_MPMB, [1, 0, 2, 3], [0, 1, 2, 3]))
    Cov_MPMB[diag_mask] = Cov_MPMB[diag_mask] / 2

    logging.info("    ...done.\n")

    return Cov_MPMB, rg_out, az_out, opt_str.subs_F_r, opt_str.subs_F_a


def MPMBCovarianceEstimationSSF(D, opt_str, R, look_angles, ground_slope):
    """MPMBCovarianceEstimation

    INPUTS:
        D: list of lists with dimensions [Num_pol x 1][Nrg x Naz x Num_Baselines]
        R: [Nrg x Naz x Nimm] distances (to bring to baseband)
        look_angles: [Nrg x Naz x Nimm] offnadir angles
        ground_slope: [Nrg x Naz] ground slope
        opt_str:
            NWmono_r: One-sided range average window [pixel]
            subs_F_r: Range subsampling factor
            NWmono_a: One-sided azimuth average window [pixel]
            subs_F_a: Azimuth subsampling factor
            polarimetric_mask: (optional) [Npol x Npol] logical matrix
                                          stating the polarization combination to
                                          be computed
            B: signal bandwidth
            dt: sampling interval
            f0: central frequency
            lambda: wavelength

    OUTPUTS:
        Cov_MPMB: Multi-polarimetric multi-baseline covariance matrix
        rg_out: range subsampled axis
        az_out: azimuth subsampled axis
    """

    acq_names = list(D.keys())
    N_imm = len(acq_names)
    first_acq_dict = D[acq_names[0]]
    pol_names = list(first_acq_dict.keys())
    num_pols = len(pol_names)
    N_rg, N_az = first_acq_dict[pol_names[0]].shape
    cell_data = True
    double_data = first_acq_dict[pol_names[0]].dtype == "complex128"

    try:
        Mask = opt_str.polarimetric_mask
    except AttributeError:
        Mask = np.ones((num_pols, num_pols))

    # Range filter matrix
    class filtering_matrix_opt_str:
        pass

    filtering_matrix_opt_str.Nin = N_rg
    filtering_matrix_opt_str.NWmono = opt_str.NWmono_r
    filtering_matrix_opt_str.subs_F = opt_str.subs_F_r
    Fr, rg_out, Rnorm = build_filtering_matrix(filtering_matrix_opt_str)
    N_rg_out = rg_out.size
    Fr_normalized = Rnorm @ Fr

    # Azimuth filter matrix
    filtering_matrix_opt_str.Nin = N_az
    filtering_matrix_opt_str.NWmono = opt_str.NWmono_a
    filtering_matrix_opt_str.subs_F = opt_str.subs_F_a
    Fa, az_out, Anorm = build_filtering_matrix(filtering_matrix_opt_str)
    N_az_out = az_out.size
    Fa_normalized_transposed = (Anorm @ Fa).T

    # Init
    Cov_MPMB = np.zeros((num_pols * N_imm, num_pols * N_imm, N_rg_out, N_az_out), dtype=np.complex64)

    class ssf_opt_str:
        pass

    ssf_opt_str.B = opt_str.B
    ssf_opt_str.f0 = opt_str.f0
    ssf_opt_str.dt = opt_str.dt

    logging.info("    MPMB covariance estimation (with spectral shift enabled)...")
    nan_mask = np.zeros((N_rg, N_az, N_imm), dtype=bool)

    for pol_name_curr in pol_names:
        all_acq_data = np.zeros((N_rg, N_az, N_imm), dtype=complex)
        for acq_idx_curr, acq_curr in enumerate(acq_names):
            all_acq_data[:, :, acq_idx_curr] = D[acq_curr][pol_name_curr]
        nan_mask = nan_mask + np.isnan(all_acq_data)

    # this has been moved from the innner cycle (referring to PM9)
    for acq_idx_curr, acq_curr in enumerate(acq_names):
        for pol_name_curr in pol_names:
            D[acq_curr][pol_name_curr][nan_mask[:, :, acq_idx_curr]] = 0

    for ch_i, pol_id_i in enumerate(pol_names):

        ind_i = np.arange(N_imm) + ch_i * N_imm

        for ch_j in np.arange(ch_i, num_pols):
            pol_id_j = pol_names[ch_j]
            if Mask[ch_i, ch_j] == 1:
                print("    .")
                ind_j = np.arange(N_imm) + ch_j * N_imm
                I = np.zeros((N_imm, N_imm, N_rg_out, N_az_out), dtype=np.complex64)
                for n_idx, acq_id_n in enumerate(acq_names):
                    if ch_i == ch_j:
                        m_min = n_idx  # ?
                    else:
                        m_min = 0

                    for m_idx in np.arange(m_min, N_imm):
                        acq_id_m = acq_names[m_idx]

                        if cell_data:
                            if double_data:
                                I1 = D[acq_id_n][pol_id_i]
                                I2 = D[acq_id_m][pol_id_j]
                            else:
                                I1 = D[acq_id_n][pol_id_i].astype("complex128")
                                I2 = (D[acq_id_m][pol_id_j]).astype("complex128")
                        else:
                            if double_data:
                                I1 = D[acq_id_n][pol_id_i]
                                I2 = D[acq_id_m][pol_id_j]
                            else:
                                I1 = D[acq_id_n][pol_id_i].astype("complex128")
                                I2 = (D[acq_id_m][pol_id_j]).astype("complex128")

                    # Demodulation to baseband
                    I1 = I1 * np.exp(-1j * 4 * np.pi / opt_str.wavelenght * R[acq_id_n])
                    I2 = I2 * np.exp(-1j * 4 * np.pi / opt_str.wavelenght * R[acq_id_m])

                    # Spectral shift filtering
                    offnadir = np.zeros((look_angles[acq_id_m].shape[0], look_angles[acq_id_m].shape[1], 2))
                    offnadir[:, :, 0] = look_angles[acq_id_n]
                    offnadir[:, :, 1] = look_angles[acq_id_m]
                    [I1, I2] = SpectralShiftFiltering(I1, I2, offnadir, ground_slope, ssf_opt_str)

                    # Modulation
                    I1 = I1 * np.exp(1j * 4 * np.pi / opt_str.wavelenght * R[acq_id_n])
                    I2 = I2 * np.exp(1j * 4 * np.pi / opt_str.wavelenght * R[acq_id_m])

                    temp = Fr_normalized @ (I1 * np.conjugate(I2))
                    temp = temp @ Fa_normalized_transposed
                    I[n_idx, m_idx, :, :] = temp.astype("complex64")

                Cov_MPMB[ind_i[:, np.newaxis], ind_j[np.newaxis, :], :, :] = I

    # Symmetric part generation
    diag_mask = np.tile(
        (np.eye(num_pols * N_imm).reshape(num_pols * N_imm, num_pols * N_imm, 1, 1)) > 0, [1, 1, N_rg_out, N_az_out],
    )
    Cov_MPMB = Cov_MPMB + np.conjugate(np.moveaxis(Cov_MPMB, [1, 0, 2, 3], [0, 1, 2, 3]))
    Cov_MPMB[diag_mask] = Cov_MPMB[diag_mask] / 2

    print("    done.\n")

    return Cov_MPMB, rg_out, az_out, opt_str.subs_F_r, opt_str.subs_F_a


def SpectralShiftFiltering(I1, I2, look_angles, ground_slope, opt_str):
    """

    INPUT
        I1: [Nr x Nc] baseband SLC
        I2: [Nr x Nc] baseband SLC
        look_angles: [Nr x Nc x 2] look angles associated with each SLC
        ground_slope: [Nr x Nc] ground slope in range direction
        opt_str.
              dt: sampling interval
              f0: central frequency
              B: bandwidth

    OUTPUT
        I1_filt: [Nr x Nc] baseband spectral shift filtered SLC
        I2_filt: [Nr x Nc] baseband spectral shift filtered SLC"""

    # Spectral shift (Hz)
    Df = opt_str.f0 * np.squeeze(np.diff(look_angles, axis=2)) / np.tan(np.mean(look_angles, axis=2) - ground_slope)

    # Space-varying demodulation phases (half)
    phi_half = np.pi * np.cumsum(Df * opt_str.dt, axis=0)

    # Demodulation
    I1_filt = I1 * np.exp(-1j * phi_half)
    I2_filt = I2 * np.exp(1j * phi_half)

    # Building filter
    if not (hasattr(opt_str, "Nfilter")):
        opt_str.Nfilter = int(np.round(I1.shape[0] / 4))

    curr_filter = firwin(numtaps=opt_str.Nfilter, cutoff=opt_str.B * opt_str.dt)
    curr_filter = curr_filter / np.sqrt(np.sum(np.abs(curr_filter) ** 2))
    curr_filter = curr_filter.reshape(opt_str.Nfilter, 1)

    # Filtering
    I1_filt = oaconvolve(I1_filt, curr_filter, mode="same", axes=0)
    I2_filt = oaconvolve(I2_filt, curr_filter, mode="same", axes=0)

    # Back to baseband
    I1_filt = I1_filt * np.exp(1j * phi_half)
    I2_filt = I2_filt * np.exp(-1j * phi_half)

    return I1_filt, I2_filt


def MPMBshuffle(MPMB_correlation, rg_vec_subs, az_vec_subs, Npol, N):

    Nrg_subs = rg_vec_subs.size
    Naz_subs = az_vec_subs.size
    MBMP_correlation = np.zeros((Npol * N, Npol * N, Nrg_subs, Naz_subs), dtype=np.complex64)
    r = MPMB_correlation.shape[0]
    I = np.eye(r)  # Identity matrix
    S = I * 0
    for i in np.arange(N):
        S[i * Npol : Npol + i * Npol, :] = I[i:r:N, :]

    for r in np.arange(Nrg_subs):
        for a in np.arange(Naz_subs):
            MBMP_correlation[:, :, r, a] = S @ MPMB_correlation[:, :, r, a] @ S.T

    return MBMP_correlation


def Covariance4D2Correlation4D(Cov_MPMB):
    """It normalizes each element of the 4-D Multi-Polarimetric Multi-Baseline
     with respect to the corresponding diagonal terms.
    
     INPUT
          Cov_MPMB: [Nimm*Npol x Nimm*Npol x Nr x Na] Covariance matrices
    
     OUTPUT
           Corr_MPMB: [Nimm*Npol x Nimm*Npol x Nr x Na] normalized covariance
                      matrices
           varargout{1}: [Nr x Na x Nimm*Npol] diagonal entries of Cov_MPMB"""
    temp, N, Nr, Na = Cov_MPMB.shape
    diag_mask = np.tile((np.eye(N).reshape(N, N, 1, 1)) > 0, [1, 1, Nr, Na])
    norm_util = Cov_MPMB[diag_mask]
    norm_diag = np.moveaxis(norm_util.reshape((N, Nr, Na)), [1, 2, 0], [0, 1, 2])
    norm_util = 1 / np.sqrt(norm_util + np.spacing(1))

    norm_util = norm_util.reshape((N, 1, Nr, Na)) * norm_util.reshape((1, N, Nr, Na))
    Corr_MPMB = norm_util * Cov_MPMB

    return Corr_MPMB, norm_diag


def Covariance2D2Correlation2D(Cov2D):
    """It normalizes each element of the covariance matrix with respect to the
     corresponding elements on the main diagonal.
    
     INPUT
          Cov2D: [N x N] covariance matrix
    
     OUTPUT
           Corr2D: [N x N] normalized covariance matrix"""
    temp = np.diag(1 / np.sqrt(np.diag(Cov2D)))
    Corr2D = temp @ Cov2D @ temp

    return Corr2D


def build_filtering_matrix(opt_str):
    """It builds a sparse matrix to carry out one-dimensional moving average.
    It deals with regularly sampled data.

    INPUTS:
       opt_str:
            Nin:    number of rows of the matrix to be filtered
            NWmono: one-sided length of the average window
            subs_F: subsampling factor of the filtered signal

     OUTPUTS:
        F:     sparse filtering matrix
        xout:  axis of the filtered signal
        Rnorm: normalizing matrix
    """

    xin = np.arange(opt_str.Nin)
    Nxin = xin.size
    xout = np.arange(0, opt_str.Nin, opt_str.subs_F)
    Nxout = xout.size

    col = np.kron(xout, np.ones((1, 2 * opt_str.NWmono + 1))) + np.kron(
        np.ones((1, Nxout)), np.arange(-opt_str.NWmono, opt_str.NWmono + 1)
    )
    row = np.kron(np.arange(Nxout), np.ones((1, 2 * opt_str.NWmono + 1)))

    ok_mask = (col >= 0) * (col < Nxin)
    Nok_mask = np.sum(ok_mask)

    F = csr_matrix((np.ones((1, Nok_mask)).flatten(), (row[ok_mask], col[ok_mask])), shape=(Nxout, Nxin))

    Rnorm = csr_matrix((1 / np.sum(F.toarray(), axis=1), (np.arange(Nxout), np.arange(Nxout))))

    return F, xout, Rnorm


def covariance_matrix_mat2vec(matrix_in):
    """covariance matrix has herimtian symmetry, so the elements under the
    diagonal are not saved in the matrix_out
    matrix_in:  [3x3xNrgxNaz] or [3x3]

    matrix_out: [6xNrgxNaz] or [6,]

    Elements in matrix in are stored in the optput one with following indexing convention:

    Those are the 3x3 matrix_in elements (for each Nrg, Naz) indexes of extraction:
    [ 0  1  2
      /  4  5
      /  /  8 ]

    Those are the 6 matrix_out elements (for each Nrg, Naz)  indexes of insertion
    [ 0 1 2 4 5 8 ]
    """
    extraction_indexes = [0, 1, 2, 4, 5, 8]

    shapes = matrix_in.shape
    if len(shapes) == 4:
        Nrg = shapes[2]
        Naz = shapes[3]

        matrix_out = np.zeros((6, Nrg, Naz), dtype=matrix_in.dtype)

        for rg_idx in np.arange(Nrg):
            for az_idx in np.arange(Naz):

                matrix_out[:, rg_idx, az_idx] = matrix_in[:, :, rg_idx, az_idx].reshape((9))[extraction_indexes]
    else:

        matrix_out = matrix_in.reshape((9))[extraction_indexes]

    return matrix_out


def covariance_matrix_vec2mat(vector_in):
    # vector_in has dimension 6
    # matrix_out has dimensions 3x3
    L = len(vector_in.shape)

    #        [ 0         1     2
    #          cj(1)     3     4
    #          cj(2)  cj(4)    5 ]
    if L == 3:

        _, Nrg, Naz = vector_in.shape
        matrix_out = np.zeros((3, 3, Nrg, Naz), dtype=vector_in.dtype)

        for rg_idx in np.arange(Nrg):
            for az_idx in np.arange(Naz):
                matrix_out[:, :, rg_idx, az_idx] = [
                    [vector_in[0, rg_idx, az_idx], vector_in[1, rg_idx, az_idx], vector_in[2, rg_idx, az_idx],],
                    [
                        np.conj(vector_in[1, rg_idx, az_idx]),
                        vector_in[3, rg_idx, az_idx],
                        vector_in[4, rg_idx, az_idx],
                    ],
                    [
                        np.conj(vector_in[2, rg_idx, az_idx]),
                        np.conj(vector_in[4, rg_idx, az_idx]),
                        vector_in[5, rg_idx, az_idx],
                    ],
                ]
    elif L == 1:

        matrix_out = np.zeros((3, 3), dtype=vector_in.dtype)

        matrix_out[:, :] = [
            [vector_in[0], vector_in[1], vector_in[2]],
            [np.conj(vector_in[1]), vector_in[3], vector_in[4]],
            [np.conj(vector_in[2]), np.conj(vector_in[4]), vector_in[5]],
        ]
    return matrix_out


def SKPD_processing(Cov_MPMB, opt_str):
    """SKPD_processing
    
     Kernel of the Sum of Kronecker Product Decomposition.
    
     INPUT
          Cov_MPMB: [N*Npol x N*Npol] covariance matrix
           opt_str.
                            N: number of images
                         Npol: number of polarizations
                   Nsubspaces: number of desired subspaces
                       Nparam: number of parameters sampling the "a" and "b"
                               range of physical admissibility
    
     OUTPUT
           out_str.
                   error: [1 x 4] logical vector.
                          1. Cov_MPMB has nan of inf values
                          2. Singular matrices have nan or inf values
                          3. Singular matrices are not full rank
                          4. Could not determine "a" or "b" ranges
                   lambda_SVD: singular values of the SKPD
                   Ro: tomographic part of the decomposition (normalized wrt
                       Ro{k}(1, 1))
                   Co: polarimetric part of the decomposition
                   ax: vector gathering the admissible values for parameter a
                   bx: vector gathering the admissible values for parameter b
                   Ra: corresponding tomographic covariance matrices (at the
                       edges of the possible range of values of "a")
                   Rb: corresponding tomographic covariance matrices (at the
                       edges of the possible range of values of "b")
                   Ca: corresponding polarimetric covariance matrices (at the
                       edges of the possible range of values of "a")
                   Cb: corresponding polarimetric covariance matrices (at the
                       edges of the possible range of values of "b")
                   full_rank_Ra: true if Ra is full rank
                   full_rank_Rb: true if Rb is full rank
                   full_rank_Ca: true if Ca is full rank
                   full_rank_Cb: true if Cb is full rank
                   Ra_coherence: tomographic coherence of Ra
                   Rb_coherence: tomographic coherence of Rb
                   Ca_coherence: tomographic coherence of Ca
                   Cb_coherence: tomographic coherence of Cb
                   Ra_more_coherent_than_Rb: [1 x 1] logical
                   ax_R_max_coh: [1 x 2] logical. True if the parameter edge
                                 provides greater coherence
                   bx_R_max_coh: [1 x 2] logical. True if the parameter edge
                                 provides greater coherence
                   R_most_coeherent: most coherent R
                   Rcoh_thin: Most coherent scattering mechanism, most
                              coherent among the range of admissibility
                   Rcoh_fat: Most coherent scattering mechanism, least
                              coherent among the range of admissibility
                   Rincoh_thin: Least coherent scattering mechanism, most
                              coherent among the range of admissibility
                   Rincoh_fat: Least coherent scattering mechanism, least
                              coherent among the range of admissibility
    
     DEPENDENCIES
                 LowRankKronApproximation
                 SemiDefinitePositivenessJointRange
                                 JointDiagonalization"""

    class out_str:
        pass

    out_str.error = np.zeros((4, 1))

    if np.any(np.isnan(Cov_MPMB) | np.isinf(Cov_MPMB)):
        out_str.error[0] = 1
        return out_str

    class LRKA_str:
        pass

    LRKA_str.A = Cov_MPMB
    LRKA_str.rank = opt_str.Nsubspaces
    LRKA_str.Nr1 = opt_str.num_pols
    LRKA_str.Nc1 = opt_str.num_pols
    LRKA_str.Nr2 = opt_str.num_acq
    LRKA_str.Nc2 = opt_str.num_acq
    LRKA_out_str = LowRankKronApproximation(LRKA_str)
    Co = LRKA_out_str.B
    Ro = LRKA_out_str.C
    lambda_SVD = LRKA_out_str.lambda_SVD

    # Demanding power to polarimetric signature
    for k in np.arange(opt_str.Nsubspaces):
        temp = Ro[k][0, 0]
        Ro[k] = Ro[k] / temp
        Co[k] = Co[k] * temp * lambda_SVD[k]

    out_str.lambda_SVD = lambda_SVD[np.arange(opt_str.Nsubspaces)]
    out_str.Ro = Ro
    out_str.Co = Co

    if np.any(np.isnan(Ro)) or np.any(np.isnan(Co)) | np.any(np.isinf(Ro)) or np.any(np.isinf(Co)):
        out_str.error[1] = 1
        return out_str

    if (
        not (np.linalg.matrix_rank(Ro[0]) == opt_str.num_acq)
        or not (np.linalg.matrix_rank(Ro[1]) == opt_str.num_acq)
        or not (np.linalg.matrix_rank(Co[0]) == opt_str.num_pols)
        or not (np.linalg.matrix_rank(Co[1]) == opt_str.num_pols)
    ):
        out_str.error[2] = 1
        return out_str

    class SDP_opt_str:
        pass

    SDP_opt_str.Ro = Ro
    SDP_opt_str.Co = Co
    SDP_opt_str.N = opt_str.Nparam
    SDP_out_str = SemiDefinitePositivenessJointRange(SDP_opt_str)
    if SDP_out_str.error:
        out_str.error[3] = 1
        return out_str
    ax = SDP_out_str.a
    bx = SDP_out_str.b

    out_str.ax = ax
    out_str.bx = bx

    out_str.Ra = list([ax[0] * Ro[0] + (1 - ax[0]) * Ro[1], ax[-1] * Ro[0] + (1 - ax[-1]) * Ro[1]])
    out_str.Rb = list([bx[0] * Ro[0] + (1 - bx[0]) * Ro[1], bx[-1] * Ro[0] + (1 - bx[-1]) * Ro[1]])

    out_str.Ca = list([-(1 - ax[0]) * Co[0] + ax[0] * Co[1], -(1 - ax[-1]) * Co[0] + ax[-1] * Co[1]])
    out_str.Cb = list([(1 - bx[0]) * Co[0] - bx[0] * Co[1], (1 - bx[-1]) * Co[0] - bx[-1] * Co[1]])

    out_str.full_rank_Ra = [np.linalg.cond(out_str.Ra[0]) < np.linalg.cond(out_str.Ra[1]), False]
    out_str.full_rank_Ra[1] = not (out_str.full_rank_Ra[0])
    out_str.full_rank_Rb = [np.linalg.cond(out_str.Rb[0]) < np.linalg.cond(out_str.Rb[1]), False]
    out_str.full_rank_Rb[1] = not (out_str.full_rank_Rb[0])

    out_str.full_rank_Ca = [np.linalg.cond(out_str.Ca[0]) < np.linalg.cond(out_str.Ca[1]), False]
    out_str.full_rank_Ca[1] = not (out_str.full_rank_Ca[0])
    out_str.full_rank_Cb = [np.linalg.cond(out_str.Cb[0]) < np.linalg.cond(out_str.Cb[1]), False]
    out_str.full_rank_Cb[1] = not (out_str.full_rank_Cb[0])

    triu_mask_R = np.triu(np.ones((opt_str.num_acq, opt_str.num_acq)), 1) > 0
    Ntriu_mask_R = (opt_str.num_acq ** 2 - opt_str.num_acq) / 2
    triu_mask_C = np.triu(np.ones((opt_str.num_pols, opt_str.num_pols)), 1) > 0
    Ntriu_mask_C = (opt_str.num_pols ** 2 - opt_str.num_pols) / 2
    out_str.Ra_coherence = np.zeros((2, 1))
    out_str.Rb_coherence = np.zeros((2, 1))
    out_str.Ca_coherence = np.zeros((2, 1))
    out_str.Cb_coherence = np.zeros((2, 1))
    for k in np.arange(2):
        O = np.diag(1 / np.sqrt(np.diag(out_str.Ra[k])))
        temp = O @ out_str.Ra[k] @ O
        out_str.Ra_coherence[k] = np.sum(np.abs(temp[triu_mask_R])) / Ntriu_mask_R
        O = np.diag(1 / np.sqrt(np.diag(out_str.Rb[k])))
        temp = O @ out_str.Rb[k] @ O
        out_str.Rb_coherence[k] = np.sum(np.abs(temp[triu_mask_R])) / Ntriu_mask_R
        O = np.diag(1 / np.sqrt(np.diag(out_str.Ca[k])))
        temp = O @ out_str.Ca[k] @ O
        out_str.Ca_coherence[k] = np.sum(np.abs(temp[triu_mask_C])) / Ntriu_mask_C
        O = np.diag(1 / np.sqrt(np.diag(out_str.Cb[k])))
        temp = O @ out_str.Cb[k] @ O
        out_str.Cb_coherence[k] = np.sum(np.abs(temp[triu_mask_C])) / Ntriu_mask_C

    out_str.Ra_more_coherent_than_Rb = np.max(out_str.Ra_coherence) > np.max(out_str.Rb_coherence)
    out_str.ax_R_max_coh = [out_str.Ra_coherence[0] > out_str.Ra_coherence[1], False]
    out_str.ax_R_max_coh[1] = not (out_str.ax_R_max_coh[0])
    out_str.bx_R_max_coh = [out_str.Rb_coherence[0] > out_str.Rb_coherence[1], False]
    out_str.bx_R_max_coh[1] = not (out_str.bx_R_max_coh[0])

    out_str.R_most_coherent = (out_str.Ra_more_coherent_than_Rb) * (
        out_str.Ra[0] * out_str.ax_R_max_coh[0] + out_str.Ra[1] * out_str.ax_R_max_coh[1]
    ) + (1 - out_str.Ra_more_coherent_than_Rb) * (
        out_str.Rb[0] * out_str.bx_R_max_coh[0] + out_str.Rb[1] * out_str.bx_R_max_coh[1]
    )

    out_str.Rcoh_thin = (out_str.Ra_more_coherent_than_Rb) * (
        out_str.Ra[0] * out_str.ax_R_max_coh[0] + out_str.Ra[1] * out_str.ax_R_max_coh[1]
    ) + (1 - out_str.Ra_more_coherent_than_Rb) * (
        out_str.Rb[0] * out_str.bx_R_max_coh[0] + out_str.Rb[1] * out_str.bx_R_max_coh[1]
    )
    out_str.Rcoh_fat = (out_str.Ra_more_coherent_than_Rb) * (
        out_str.Ra[0] * out_str.ax_R_max_coh[1] + out_str.Ra[1] * out_str.ax_R_max_coh[0]
    ) + (1 - out_str.Ra_more_coherent_than_Rb) * (
        out_str.Rb[0] * out_str.bx_R_max_coh[1] + out_str.Rb[1] * out_str.bx_R_max_coh[0]
    )
    out_str.Rincoh_thin = (1 - out_str.Ra_more_coherent_than_Rb) * (
        out_str.Ra[0] * out_str.ax_R_max_coh[0] + out_str.Ra[1] * out_str.ax_R_max_coh[1]
    ) + (out_str.Ra_more_coherent_than_Rb) * (
        out_str.Rb[0] * out_str.bx_R_max_coh[0] + out_str.Rb[1] * out_str.bx_R_max_coh[1]
    )
    out_str.Rincoh_fat = (1 - out_str.Ra_more_coherent_than_Rb) * (
        out_str.Ra[0] * out_str.ax_R_max_coh[1] + out_str.Ra[1] * out_str.ax_R_max_coh[0]
    ) + (out_str.Ra_more_coherent_than_Rb) * (
        out_str.Rb[0] * out_str.bx_R_max_coh[1] + out_str.Rb[1] * out_str.bx_R_max_coh[0]
    )

    return out_str


def LowRankKronApproximation(opt_str):
    """It returns the sum of Kronecker product decomposition of the matrix A
     just as the singular value decomposition. Matrix A is decomposed as:
     for k = 1:rank
         A = A + kron(B{k}, C{k});
     end
    
     INPUT
          opt_str.
                  A: [Nr x Nc] matrix to be decomposed and approximated
                  rank: rank of the approximation
                  Nr1: number of rows of the matrix B
                  Nc1: number of columns of the matrix B
                  Nr2: number of rows of the matrix C
                  Nr2: number of columns of the matrix C
    
     OUTPUT
           out_str.
                   lambda: singular values of the decomposition
                   B: [rank x 1] cell of [Nr1 x Nc1] matrices
                   C: [rank x 1] cell of [Nr2 x Nc2] matrices"""
    RA = np.zeros((opt_str.Nr1 * opt_str.Nc1, opt_str.Nr2 * opt_str.Nc2), dtype=np.complex128)
    for p in np.arange(opt_str.Nc1):
        Ap = np.zeros((opt_str.Nr1, opt_str.Nr2 * opt_str.Nc2), dtype=np.complex128)
        ind_p = np.arange(p * opt_str.Nc2, (p + 1) * opt_str.Nc2)
        for q in np.arange(opt_str.Nr1):
            ind_q = np.arange(q * opt_str.Nr2, (q + 1) * opt_str.Nr2)
            Aqp = opt_str.A[ind_q[:, np.newaxis], ind_p[np.newaxis, :]]
            Ap[q, :] = Aqp.transpose().flatten()
        ind = np.arange(p * opt_str.Nr1, (p + 1) * opt_str.Nr1)
        RA[ind, :] = Ap

    U, S, VH = np.linalg.svd(RA)

    class out_str:
        pass

    out_str.lambda_SVD = S

    out_str.B = list()
    out_str.C = list()
    for k in np.arange(opt_str.rank):
        out_str.B.append(np.reshape(U[:, k], (opt_str.Nr1, opt_str.Nc1), order="F"))
        out_str.C.append(np.reshape(np.transpose(VH)[:, k], (opt_str.Nr2, opt_str.Nc2), order="F"))  # warning !

    return out_str


def SemiDefinitePositivenessJointRange(opt_str):
    """Given the relationships:
                   Rg = a*Ro{1} + (1 - a)*Ro{2}
                   Rv = b*Ro{2} + (1 - b)*Ro{2}
                   Cg = (1 - b)*C{1} - b*C{2}
                   Cv = -(1 - a)*C{1} + a*C{2}
     this routine returns the range of values for parameters "a" and "b" such
     that matrices Rg, Rv, Cg, Cv are semi definite positive.
    
     INPUT
          opt_str.
                  Ro: [2 x 1] cell of [N x N] Hermitian matrices
                  Co: [2 x 1] cell of [Npol x Npol] Hermitian matrices
                   N: number of points sampling the admissible range of values
    
     OUTPUT
           out_str.
                   a: [1 x opt_str.N] vector
                   b: [1 x opt_str.N] vector
                   error: true if an error has occurred
    
     DEPENDENCIES
                 JointDiagonalization"""

    class out_str:
        pass

    out_str.error = False

    # Semi positive definiteness of the interferometric covariance matrices
    U = JointDiagonalization(opt_str.Ro[0:2])
    d1 = np.real(np.diag(U.conj().transpose() @ opt_str.Ro[0] @ U))
    d2 = np.real(np.diag(U.conj().transpose() @ opt_str.Ro[1] @ U))

    dd = d1 - d2
    indm = dd < 0
    indp = dd > 0
    temp = -d2 / dd

    try:
        ab_lo = np.max(temp[indp])
    except ValueError:
        ab_lo = -np.inf
    try:
        ab_up = np.min(temp[indm])
    except ValueError:
        ab_up = np.inf
    # Semi positive definiteness of the polarimetric covariance matrices
    U = JointDiagonalization(opt_str.Co[0:2])
    d1 = np.real(np.diag(U.conj().transpose() @ opt_str.Co[0] @ U))
    d2 = np.real(np.diag(U.conj().transpose() @ opt_str.Co[1] @ U))

    dd = d1 + d2
    indm = dd < 0
    indp = dd > 0
    temp = d1 / dd

    try:
        a_lo = np.max(temp[indp])
    except ValueError:
        a_lo = -np.inf
    try:
        a_up = np.min(temp[indm])
    except ValueError:
        a_up = np.inf

    dd = -d1 - d2
    indm = dd < 0
    indp = dd > 0
    temp = -d1 / dd

    try:
        b_lo = np.max(temp[indp])
    except ValueError:
        b_lo = -np.inf
    try:
        b_up = np.min(temp[indm])
    except ValueError:
        b_up = np.inf

    if a_lo < b_up:
        out_str.error = True
    # Semi positive definiteness of both covariance matrices
    a_lo = np.max((a_lo, ab_lo))
    a_up = np.min((a_up, ab_up))
    b_lo = np.max((b_lo, ab_lo))
    b_up = np.min((b_up, ab_up))

    Da = a_up - a_lo
    Db = b_up - b_lo
    if Da < 0 or Db < 0:
        out_str.error = True

    out_str.a = np.linspace(a_lo, a_up, opt_str.N)
    out_str.b = np.linspace(b_lo, b_up, opt_str.N)

    return out_str


def JointDiagonalization(C):
    """It finds the matrix U that diagonalizes C{1} and C{2} at the same time
     (it means that U'*C{1}*U and U'*C{2}*U are diagonal)
    
     INPUT
          C: [2 x 1] cell of [N x N] Hermitian matrices
    
     OUTPUT
           U: diagonalizing matrix"""
    A, V = np.linalg.eig(C[1] @ np.linalg.inv(C[0]))
    U = np.linalg.inv(V.conj().transpose())

    return U



# SPDX-FileCopyrightText: Aresys S.r.l. <info@aresys.it>
# SPDX-License-Identifier: MIT

"""
Binary debug data file management module
----------------------------------------
"""

import struct
import numpy as np

# The debug data file binary header is expected to be a sequence of three 32-bit integers, representing:
# * the cell type identifier
# * the number of samples of the data matrix
# * the number of lines of the data matrix
_HEADER_FORMAT = 'iii'
_HEADER_SIZE = struct.calcsize(_HEADER_FORMAT)

_CELLTYPE_TO_DTYPE_MAP = {
    0: 'float32',
    1: 'complex64',
    2: 'float64',
    3: 'complex128',
    4: 'int16',
    5: 'int8',
    6: 'int32',
    7: 'int64',
}


def read_debug(filename):
    """Read binary debug data file content.

    :param filename: binary debug data file name

    :return: a numpy array containing the data read from the input binary debug data file
    """

    with open(filename, mode='rb') as fdesc:
        header = fdesc.read(_HEADER_SIZE)

        celltype, samples, lines = struct.unpack(_HEADER_FORMAT, header)

        try:
            dtype = _CELLTYPE_TO_DTYPE_MAP[celltype]
        except KeyError as exc:
            raise TypeError('Unknown cell type (cell type id = {})'.format(celltype)) from exc

        if samples == 0 or lines == 0:
            return np.empty((lines, samples), dtype)

        data = np.fromfile(fdesc, dtype)

        return data.reshape((lines, samples))
